%
% algorithm for compute n-dimensional reachable set
%

function [R,Rsteps,Rdiff,Err] = compute_reachableset(x0, t0, T, N, mesh, bounds, dimu, rhs, rhs_u, rhs_x)

    n=size(x0,1);
    u=zeros(dimu*N,1);
    h=(T-t0)/N;
    point=zeros(n,1);

    index=ones(1,n);
    countdir=ones(1,n);
    numpoints=1;
    for i=1:n
        numpoints = numpoints*mesh(i);
    end

    eps=min((bounds(:,2)-bounds(:,1))./mesh')*1e-2;

    R=zeros(1,n);
    Rsteps=zeros(1,1);
    Rdiff=zeros(1,1);
    errpoints=0;
    Err=zeros(1,n);

    totaltime = 0.0;
    pnr=1;
    soutput = sprintf('');

    for p=1:numpoints
        t1 = rem(now,1)*24*60*60;
        for i=1:n
            point(i) = bounds(i,1) + (bounds(i,2) - bounds(i,1))*(index(i)-1)/(mesh(i)-1);
        end

        oldu = u;

        if p==1
            [u,err,steps,~,s_b,y_b,m_b] = ip(u, ...
                                                    @(u)(f(dimu, x0, t0, rhs, h, u, point)), ...
                                                    @g, ...
                                                    @(u)(f_u(dimu, x0, t0, rhs, rhs_x, rhs_u, h, u, point)), ...
                                                    @g_u, ...
                                                    2*dimu*N, eps, 1e-9, 100, 'lbfgs',1);
        else
            [u,err,steps,~,s_b,y_b,m_b] = ip(u, ...
                                                    @(u)(f(dimu, x0, t0, rhs, h, u, point)), ...
                                                    @g, ...
                                                    @(u)(f_u(dimu, x0, t0, rhs, rhs_x, rhs_u, h, u, point)), ...
                                                    @g_u, ...
                                                    2*dimu*N, eps, 1e-9, 100, 'lbfgs',1,y_b,s_b,m_b);
        end

        if err == 0

            temp_t = t0;
            temp_x = x0;
            for i=1:N
                temp_x = solveode(temp_t, temp_x, temp_t+h, rhs, u((i-1)*dimu+1:i*dimu));
                temp_t = temp_t + h;
            end
            R(pnr,:)=temp_x';
            Rsteps(pnr)=steps;
            Rdiff(pnr) = norm(u-oldu,2);
            pnr = pnr+1;
        else
            u = oldu;
            errpoints = errpoints+1;
            Err(errpoints,:) = point';
            for i=1:size(soutput,2)
                fprintf('\b');
            end
            %fprintf('IP failed...\n');
            soutput=sprintf('');
        end

        cursor = 1;
        while cursor <= n
            index(cursor) = index(cursor) + countdir(cursor);
            if countdir(cursor)==1
                if index(cursor) > mesh(cursor)
                    index(cursor) = mesh(cursor);
                    countdir(cursor) = -1;
                    cursor = cursor + 1;
                else
                    break;
                end
            else
                if index(cursor) == 0
                    index(cursor) = 1;
                    countdir(cursor) = 1;
                    cursor = cursor + 1;
                else
                    break;
                end
            end
        end
        t2 = rem(now,1)*24*60*60;
        totaltime = totaltime + (t2-t1);
        timeperstep=totaltime/p;
        remainingtime=(numpoints-p)*timeperstep;

        for i=1:size(soutput,2)
            fprintf('\b');
        end
        soutput = sprintf('%1.1f done... remaining time: %1.2f sec', p/numpoints*100.0, remainingtime);
        fprintf(soutput);
    end
    fprintf('\n');

    if errpoints > 0
        fprintf('there have been errors during computation... check "Err"-Matrix\n');
    end

end



function y = f(dimu, x0, t0, rhs, h, u, target)

    N = size(u,1)/dimu;
    n = size(x0,1);
    x = x0;
    t = t0;

    for i=1:N
        x = solveode(t, x, t+h, rhs, u((i-1)*dimu+1:i*dimu));
        t = t+h;
    end

    y = 0.0;
    for i=1:n
        y = y + (x(i) - target(i))^2;
    end 

end



function du = f_u(dimu, x0, t0, rhs, rhs_x, rhs_u, h, u, target)

    N = size(u,1)/dimu;
    n = size(x0,1);
    x = zeros(n, (dimu*N)+1);
    x(:,(dimu*N)+1) = x0;
    t = t0;

    for i=1:N
        x = solveode(t, x, t+h, @(t,x,u)(rhs_extended(t,x,u,i,rhs,rhs_x,rhs_u)), u((i-1)*dimu+1:i*dimu));
        t = t+h;
    end

    du = zeros(1,dimu*N);
    for i=1:n
        du = du + 2.0*(x(i,dimu*N+1) - target(i))*x(i,1:dimu*N);
    end

end


function y = g(u)

    global lb;
    global ub;

    N = size(u,1);
    y = zeros(2*N,1);

    for i=1:N
        y(2*(i-1)+1) = u(i) - lb;
        y(2*(i-1)+2) = ub - u(i);
    end

end

function du = g_u(u)

    N = size(u,1);

    du = zeros(2*N,N);

    for i=1:N
        du(2*(i-1)+1, i) = 1;
        du(2*(i-1)+2, i) = -1;
    end

end

function x = solveode(t0, x0, T, f, u)

    h = (T-t0);
    x = x0;
    t = t0;

    % Euler
    %x = x + h*f(t, x, u);

    % Heun

    %  k1 = f(t,x,u);
    %  k2 = f(t+h, x+h*k1, u);
    %  x = x + h/2*(k1+k2);

    % RK 4
    k1 = f(t,x,u);
    k2 = f(t+h/2, x+h*k1/2,u);
    k3 = f(t+h/2, x+h*k2/2,u);
    k4 = f(t+h, x+h*k3,u);
    x=x+h*(k1/6 + k2/3 + k3/3 + k4/6);

end

function dx = rhs_extended(t, x, u, useg, rhs, rhs_x, rhs_u)

    dimu = size(u,1);
    idx = size(x,2);

    dx = zeros(size(x,1), idx);

    dx(:,idx) = rhs(t, x(:,idx), u);
    for j=1:dimu*useg
        dx(:,j) = rhs_x(t, x(:,idx),u)*x(:,j);
    end

    dx(:,dimu*(useg-1) + 1:dimu*useg) = dx(:,dimu*(useg-1) + 1:dimu*useg) + rhs_u(t, x(:,idx), u);

end