function [R,Rsteps,Bsize] = compute_reachableset(x0, t0, T, N, mesh, bounds, fac, dimu, control_lb, control_ub, Pl, Pu, state_lb, state_ub, rhs, rhs_u, rhs_x)

    %%%%%% Method %%%%%%
    % Lobatto IIIB (order 4)
    A=[ 1/6, -1/6,   0;
        1/6,  1/3,   0;
        1/6,  5/6,   0 ];
    b=[ 1/6,  2/3, 1/6 ];
    c=[   0,  1/2,   1 ];

    stages=size(A,1);


    n=size(x0,1);
    hstep=(T-t0)/N;
    point=zeros(n,1);

    numpoints=1;
    for i=1:n
        numpoints = numpoints*mesh(i);
    end

    eps=(bounds(:,2)-bounds(:,1))./mesh';

    R=zeros(1,n);
    Rsteps=zeros(1,1);
    Bsize=zeros(1,1);

    pnr=1;
    soutput = sprintf('');

    temp_t = t0;
    temp_x = x0;
    ubuffer = zeros(dimu*N+(stages+1)*n*N,1);
    for i=1:N
        temp_x = solveode(temp_t, temp_x, temp_t+hstep, rhs, zeros(dimu,1));
        ubuffer(N*dimu+((i-1)*n+1:i*n)) = temp_x;

        for j=1:stages
            ubuffer(N*dimu+n*N+(i-1)*stages*n+(j-1)*n+(1:n)) = rhs(temp_t+hstep*c(j), temp_x, zeros(dimu,1));
        end

        temp_t = temp_t + hstep;
    end

    [neq,eq] = eval_functions(ubuffer, dimu, N, x0, t0, A, b, c, rhs, control_lb, control_ub, Pl, Pu, state_lb, state_ub, hstep, temp_x, eps/2*fac);

    idx = round((temp_x-bounds(:,1))./(bounds(:,2)-bounds(:,1)).*(mesh'-1))+1;

    idxbuffer = idx;
    sbuffer = neq;
    zbuffer = zeros(size(neq,1),1);
    ybuffer = zeros(size(eq,1),1);
    m_bbuffer = [];
    s_bbuffer = [];
    y_bbuffer = [];

    buffersize=1;
    buffermaxsize=1;
    bufferstart=1;
    bufferend=1;

    m_b=[];
    s_b=[];
    y_b=[];

    firstrun = 1;

    field=zeros(mesh(1),mesh(2));
    env=[-1 -1 -1 0 1 1  1  0 ;
         -1  0  1 1 1 0 -1 -1 ];
    envsize=size(env,2);
    envstarter=0;

    while buffersize>0 
        Bsize(end+1)=buffersize;
        %Punkt vom Puffer holen
        idx = idxbuffer(:,bufferstart);

        if field(idx(1),idx(2)) == 0
            for i=1:n
                point(i) = bounds(i,1) + (bounds(i,2) - bounds(i,1))*(idx(i)-1)/(mesh(i)-1);
            end

            % Kontrolle vom Stack holen
            u = ubuffer(:,bufferstart);
            for i=1:N*(stages+1)
                u(dimu*N+(2*i)-1:dimu*N+2*i) = u(dimu*N+(2*i)-1:dimu*N+2*i)-eps;
            end

            s = sbuffer(:,bufferstart);
            z = zbuffer(:,bufferstart);
            y = ybuffer(:,bufferstart);

            if firstrun == 0
                m_b = m_bbuffer(:,bufferstart);
                s_b = s_bbuffer(:,bufferstart);
                y_b = y_bbuffer(:,bufferstart);
            end

            xi=1e-9;

            [u,err,steps,~, s, z, y, ~, m_b, s_b, y_b,~] = ip(u, s, z, y, [], m_b, s_b, y_b, ...
                                    @(u)(eval_functions(u, dimu, N, x0, t0, A, b, c, rhs, control_lb, control_ub, Pl, Pu, state_lb, state_ub, hstep, point, eps/2*fac)), ...
                                    @(u)(eval_functions_grad(u, dimu, N, x0, t0, A, b, c, rhs, rhs_x, rhs_u, control_lb, control_ub, Pl, Pu, state_lb, state_ub, hstep, point, eps/2*fac)), ...
                                    size(zbuffer,1), ...
                                    size(ybuffer,1), ...
                                    min(eps)*1e-2, 10, xi, 100, 'lbfgs',1);
            bufferstart = bufferstart+1;
            if bufferstart > buffermaxsize
                bufferstart=1;
            end
            buffersize = buffersize - 1;

            firstrun = 0;
            if err==0
                R(pnr,:)=point';
                Rsteps(pnr)=steps;
                pnr = pnr+1;
                field(idx(1),idx(2))=1;

                for i=1:envsize
                    envidx = i+envstarter;
                    while envidx > envsize
                        envidx = envidx - envsize;
                    end
                    tempidx = idx+env(:,envidx);

                    if tempidx>[0;0]
                        if tempidx <= mesh'
                            if field(tempidx(1), tempidx(2)) == 0

                                buffersize = buffersize + 1;
                                if buffersize > buffermaxsize
                                    buffermaxsize = buffermaxsize+1;
                                    insertindex = buffermaxsize;
                                else
                                    bufferend = bufferend + 1;
                                    if bufferend > buffermaxsize
                                        bufferend = 1;
                                    end
                                    insertindex = bufferend;
                                end

                                ubuffer(:,insertindex) = u;
                                idxbuffer(:,insertindex) = tempidx;
                                sbuffer(:,insertindex) = s;
                                zbuffer(:,insertindex) = z;
                                ybuffer(:,insertindex) = y;
                                m_bbuffer(:,insertindex) = m_b;
                                s_bbuffer(:,insertindex) = s_b;
                                y_bbuffer(:,insertindex) = y_b;
                            end
                        end
                    end
                end
                envstarter=envstarter+1;
            end
        else
            bufferstart = bufferstart+1;
            if bufferstart > buffermaxsize
                bufferstart=1;
            end
            buffersize = buffersize - 1;
        end

        for i=1:size(soutput,2)
            fprintf('\b');
        end
        soutput = sprintf('buffersize=%d (max %d), pnr=%d, steps=%d', buffersize, buffermaxsize, pnr,steps);
        fprintf(soutput);



    end
    fprintf('\n');

end



function x = solveode(t0, x0, T, f, u)

    h = (T-t0);
    x = x0;
    t = t0;

    % Euler
    %x = x + h*f(t, x, u);

    % Heun

    %  k1 = f(t,x,u);
    %  k2 = f(t+h, x+h*k1, u);
    %  x = x + h/2*(k1+k2);

    %RK 4

    k1 = f(t,x,u);
    k2 = f(t+h/2, x+h*k1/2,u);
    k3 = f(t+h/2, x+h*k2/2,u);
    k4 = f(t+h, x+h*k3,u);
    x=x+h*(k1/6 + k2/3 + k3/3 + k4/6);

end

