#include "scheduler.h"

#include <list>
#include <set>
#include <map>
#include <iostream>
#include <cmath>
#include <cstring>
#include "quotientgraph.h"
#include "permutation.h"

#include <cuda_runtime.h>

using namespace std;

//#define VERBOSE

namespace OptChol
{

int * Scheduler::generate(int n, double * A)
{
	int counter;

	int * packedinfo;

	int * n_rhselim = new int[n-1];
	int * n_matrixelim = new int[n-1];
	int * n_backsubst = new int[n-1];
	int * n_nonzeros = new int[n];
	int * sdiag = new int[n];
	int * mapped_sdiag = new int[n];
	int * idxmapping = new int[n];
	list<int> rhselim;
	list<int> mapped_rhselim;
	list<int> backsubst;
	list<int> matrixelim;
	list<int> mapped_matrixelim;
	list<int> nonzeros;

	set<unsigned int> matrix_nonzeros;

	for (int j=0; j<n; j++)
	{
		for (int i=0; i<n; i++)
		{
			if (abs(*A) != 0.0)
			{
				matrix_nonzeros.insert(OptChol::Quotientgraph::coord_encode(i,j));
#ifdef VERBOSE				
				cout << "1 ";
			} else
			{
				cout << "0 ";
#endif
			}
			A++;
		}

#ifdef VERBOSE				
		cout << endl;
#endif
	}

	OptChol::Permutation::calc(n, matrix_nonzeros, sdiag);
#ifdef VERBOSE				
	cout << "PERMUTATION: ";
#endif
	for (int i=0; i<n; i++)
	{
		sdiag[i]++; // Fortran idx
//		sdiag[i] = i+1;
#ifdef VERBOSE				
		cout << sdiag[i] << " ";
#endif
	}
//	if (n == 39)
//	{
//		sdiag[1] = 10+1;
//		sdiag[5] = 5+1;
//		sdiag[10] = 15+1;
//		sdiag[15] = 1+1;
//	}
#ifdef VERBOSE				
	cout << endl;
#endif

	int globalcounter=0;

	for (int i=0; i<n;i++)
	{
#ifdef VERBOSE				
		cout << "Dec Schritt " << i << endl;
#endif

		typedef struct {
			int j;
			int k;
		} T_MATELIM;

		set<int> colnz;
		map<int, T_MATELIM> ordered_matrixelim;
		set<int> ordered_rhselim;


		int i2 = sdiag[i]-1;
		colnz.insert(i2);
   		matrix_nonzeros.insert(OptChol::Quotientgraph::coord_encode(i2,i2));
      
		if (i < n-1)
		{
			counter = 0;
			for (int j=i+1; j<n; j++)
			{
				int j2 = sdiag[j]-1;
				int i2 = sdiag[i]-1;
				if (matrix_nonzeros.find(OptChol::Quotientgraph::coord_encode(j2,i2)) != matrix_nonzeros.end())
				{
//					cout << "  b(" << j2 << ") -= b(" << i2 << ")*g(" << j2 << "," << i2 << ")" << endl;
					ordered_rhselim.insert(j2);
					counter++;
				}

			}
			n_rhselim[i] = counter;

			counter = 0;
			for (int j=i+1; j <n; j++)
			{
				for (int k=i+1; k<=j; k++)
				{
					int i2 = sdiag[i]-1;
					int j2 = sdiag[j]-1;
					int k2 = sdiag[k]-1;
					if (matrix_nonzeros.find(OptChol::Quotientgraph::coord_encode(j2,i2)) != matrix_nonzeros.end() && matrix_nonzeros.find(OptChol::Quotientgraph::coord_encode(k2,i2)) != matrix_nonzeros.end())
					{
						if (matrix_nonzeros.find(OptChol::Quotientgraph::coord_encode(j2,k2)) == matrix_nonzeros.end())
						{
							matrix_nonzeros.insert(OptChol::Quotientgraph::coord_encode(j2,k2));
#ifdef VERBOSE				
							cout << "fillin: g(" << j2 << "," << k2 << ")" << endl;
#endif
						}
//						cout << "  g(" << j2 << "," << k2 << ") -= g(" << j2 << "," << i2 << ")*g(" << k2 << "," << i2 << ")" << endl;
						T_MATELIM item;
						item.j = j2;
						item.k = k2;
						ordered_matrixelim.insert(pair<int,T_MATELIM>((k2<<16)+j2, item));
						colnz.insert(j2);
						colnz.insert(k2);
						counter++;
					}
					globalcounter++;
				}
			}
			n_matrixelim[i] = counter;
      	}
      	n_nonzeros[i] = colnz.size();
#ifdef VERBOSE
		cout << "Nonzeros: ";
#endif
		for (int i=0; i<n; i++) idxmapping[i] = -1;
		int cnt=0;
		for (set<int>::iterator it=colnz.begin(); it != colnz.end(); it++)
		{
			nonzeros.push_back(*it);
			idxmapping[*it] = cnt++;
#ifdef VERBOSE
			cout << *it << " ";
#endif
		}
		mapped_sdiag[i] = idxmapping[i2];
#ifdef VERBOSE
		cout << endl;
#endif
#ifdef VERBOSE				
		cout << "  g([" << mapped_sdiag[i] << "],[" << mapped_sdiag[i] << "])" << endl;
#endif
		for (set<int>::iterator it=ordered_rhselim.begin(); it != ordered_rhselim.end(); it++)
		{
			rhselim.push_back( *it );
			if (idxmapping[*it] == -1)
			{
				cout << "FATAL ERROR! Chol-Scheduler idx mapping = -1!" << endl;
				throw 1;
			}
			mapped_rhselim.push_back(idxmapping[*it]);
#ifdef VERBOSE
			cout << "  b([" << idxmapping[*it] << "]) -= b(" << i2 << ")*g([" << idxmapping[*it] << "]," << i2 << ")" << endl;
#endif
		}
		for (map<int,T_MATELIM>::iterator it=ordered_matrixelim.begin(); it != ordered_matrixelim.end(); it++)
		{
			T_MATELIM & item = it->second;
			matrixelim.push_back( item.j );
			matrixelim.push_back( item.k );
			if (idxmapping[item.j] == -1 || idxmapping[item.k] == -1)
			{
				cout << "FATAL ERROR! Chol-Scheduler idx mapping = -1!" << endl;
				throw 1;
			}
			mapped_matrixelim.push_back(idxmapping[item.j]);
			mapped_matrixelim.push_back(idxmapping[item.k]);
#ifdef VERBOSE
			cout << "  g(" << item.j << "," << item.k << ") -= g([" << idxmapping[item.j] << "]," << i2 << ")*g([" << idxmapping[item.k] << "]," << i2 << ")" << endl;
#endif
		}
   }

   for (int i=0; i<n-1; i++)
   {
#ifdef VERBOSE				
		cout << "BS Schritt " << i << endl;
#endif

		set<int> ordered_backsubst;

		counter = 0;
		int bidx = n-i-1;
		int bidx2 = sdiag[bidx]-1;
		for (int j=0; j<bidx; j++)
		{
			int j2 = sdiag[j]-1;
			if (matrix_nonzeros.find(OptChol::Quotientgraph::coord_encode(bidx2,j2)) != matrix_nonzeros.end())
			{
				ordered_backsubst.insert(j2);
				counter++;
			}
		}

		n_backsubst[i] = counter;

		for (set<int>::iterator it=ordered_backsubst.begin(); it != ordered_backsubst.end(); it++)
		{
			backsubst.push_back( *it );
#ifdef VERBOSE
			cout << "  b(" << *it << ") -= b(" << bidx2 << ")*g(" << bidx2 << "," << *it << ")" << endl;
#endif
		}
   	}

#ifdef VERBOSE				
   	cout << "Alle eliminationen: " << globalcounter << endl;
   	cout << "Sparse: " << matrixelim.size()/2 << endl;
#endif

	int totalsize = 12 + n + (n-1) + (n-1) + (n-1) + n + rhselim.size() + matrixelim.size() + backsubst.size() + nonzeros.size() + mapped_rhselim.size() + mapped_matrixelim.size() + n;
	packedinfo = new int[totalsize];
	packedinfo[0] = 13;
	packedinfo[1] = packedinfo[0]+n;
	packedinfo[2] = packedinfo[1]+(n-1);
	packedinfo[3] = packedinfo[2]+(n-1);
	packedinfo[4] = packedinfo[3]+(n-1);
	packedinfo[5] = packedinfo[4]+n;
	packedinfo[6] = packedinfo[5]+rhselim.size();
	packedinfo[7] = packedinfo[6]+matrixelim.size();
	packedinfo[8] = packedinfo[7]+backsubst.size();
	packedinfo[9] = packedinfo[8]+nonzeros.size();
	packedinfo[10] = packedinfo[9]+mapped_rhselim.size();
	packedinfo[11] = packedinfo[10]+mapped_matrixelim.size();
	memcpy(packedinfo+packedinfo[0]-1, sdiag, sizeof(int)*n);
	memcpy(packedinfo+packedinfo[1]-1, n_rhselim, sizeof(int)*(n-1));
	memcpy(packedinfo+packedinfo[2]-1, n_matrixelim, sizeof(int)*(n-1));
	memcpy(packedinfo+packedinfo[3]-1, n_backsubst, sizeof(int)*(n-1));
	memcpy(packedinfo+packedinfo[4]-1, n_nonzeros, sizeof(int)*n);
	memcpy(packedinfo+packedinfo[11]-1, mapped_sdiag, sizeof(int)*n);

	int * ptr = packedinfo+packedinfo[5]-1;
	for (list<int>::iterator it = rhselim.begin(); it != rhselim.end(); ++it)
	{
		*ptr = (*it)+1;
		ptr++;
	}

	for (list<int>::iterator it = matrixelim.begin(); it != matrixelim.end(); ++it)
	{
		*ptr = (*it)+1;
		ptr++;
	}

	for (list<int>::iterator it = backsubst.begin(); it != backsubst.end(); ++it)
	{
		*ptr = (*it)+1;
		ptr++;
	}

	for (list<int>::iterator it = nonzeros.begin(); it != nonzeros.end(); ++it)
	{
		*ptr = (*it);
		ptr++;
	}

	for (list<int>::iterator it = mapped_rhselim.begin(); it != mapped_rhselim.end(); ++it)
	{
		*ptr = (*it);
		ptr++;
	}

	for (list<int>::iterator it = mapped_matrixelim.begin(); it != mapped_matrixelim.end(); ++it)
	{
		*ptr = (*it);
		ptr++;
	}

	delete[] n_rhselim;
	delete[] n_matrixelim;
	delete[] n_backsubst;
	delete[] n_nonzeros;
	delete[] sdiag;
	delete[] mapped_sdiag;
	delete[] idxmapping;

	return packedinfo;
}


void Scheduler::createCUDAData(int n, int * packedinfo, T_CUDADATA & cudadata)
{
	cudadata.n=n;

	int * n_rhselim = packedinfo+packedinfo[1]-1;
	int * n_matrixelim = packedinfo+packedinfo[2]-1;
	int * n_backsubst = packedinfo+packedinfo[3]-1;
	int * n_nonzeros = packedinfo+packedinfo[4]-1;

	cudaMalloc(&(cudadata.sdiag), sizeof(int)*n);
	cudaMalloc(&(cudadata.mapped_sdiag), sizeof(int)*n);
	cudaMalloc(&(cudadata.n_rhselim), sizeof(int)*(n-1));
	cudaMalloc(&(cudadata.n_matrixelim), sizeof(int)*(n-1));
	cudaMalloc(&(cudadata.n_backsubst), sizeof(int)*(n-1));
	cudaMalloc(&(cudadata.n_nonzeros), sizeof(int)*n);

	cudaMemcpy(cudadata.sdiag, packedinfo+packedinfo[0]-1, sizeof(int)*n, cudaMemcpyHostToDevice);
	cudaMemcpy(cudadata.mapped_sdiag, packedinfo+packedinfo[11]-1, sizeof(int)*n, cudaMemcpyHostToDevice);
	cudaMemcpy(cudadata.n_rhselim, n_rhselim, sizeof(int)*(n-1), cudaMemcpyHostToDevice);
	cudaMemcpy(cudadata.n_matrixelim, n_matrixelim, sizeof(int)*(n-1), cudaMemcpyHostToDevice);
	cudaMemcpy(cudadata.n_backsubst, n_backsubst, sizeof(int)*(n-1), cudaMemcpyHostToDevice);
	cudaMemcpy(cudadata.n_nonzeros, n_nonzeros, sizeof(int)*n, cudaMemcpyHostToDevice);

	int counter;

	counter=0;
	for (int i=0; i<n; i++)
	{
		counter += n_nonzeros[i];
	}
	cudadata.total_nonzeros = counter;
	cudaMalloc(&(cudadata.nonzeros), sizeof(int)*counter);
	cudaMemcpy(cudadata.nonzeros, packedinfo+packedinfo[8]-1, sizeof(int)*counter, cudaMemcpyHostToDevice);

	counter=0;
	for (int i=0; i<n-1; i++)
	{
		counter += n_rhselim[i];
	}
	cudadata.total_rhselim = counter;
	cudaMalloc(&(cudadata.mapped_rhselim), sizeof(int)*counter);
	cudaMemcpy(cudadata.mapped_rhselim, packedinfo+packedinfo[9]-1, sizeof(int)*counter, cudaMemcpyHostToDevice);
	cudaMalloc(&(cudadata.rhselim), sizeof(int)*counter);
	cudaMemcpy(cudadata.rhselim, packedinfo+packedinfo[5]-1, sizeof(int)*counter, cudaMemcpyHostToDevice);

	counter=0;
	for (int i=0; i<n-1; i++)
	{
		counter += n_matrixelim[i]*2;
	}
	cudadata.total_matrixelim = counter;
	cudaMalloc(&(cudadata.mapped_matrixelim), sizeof(int)*counter);
	cudaMemcpy(cudadata.mapped_matrixelim, packedinfo+packedinfo[10]-1, sizeof(int)*counter, cudaMemcpyHostToDevice);

	counter=0;
	for (int i=0; i<n-1; i++)
	{
		counter += n_backsubst[i];
	}
	cudadata.total_backsubst = counter;
	cudaMalloc(&(cudadata.backsubst), sizeof(int)*counter);
	cudaMemcpy(cudadata.backsubst, packedinfo+packedinfo[7]-1, sizeof(int)*counter, cudaMemcpyHostToDevice);

	cudadata.threadsperstid = 32;
	cudadata.problemsperblock = max(min<int>((6*1024 - n*sizeof(int))/(n*sizeof(double)+2*sizeof(int)), 192/cudadata.threadsperstid),1);
	cudadata.sharedmemsize = n*(sizeof(int)+cudadata.problemsperblock*sizeof(double))+2*cudadata.problemsperblock*sizeof(int);

	cudadata.trans_rows = min(16,n);
	cudadata.trans_problemsperblock = 256/(cudadata.trans_rows*cudadata.trans_rows);
	cudadata.trans_sharedmemsize = ((cudadata.trans_rows%2==0) ?
		(cudadata.trans_problemsperblock*(cudadata.trans_rows+1)*cudadata.trans_rows) :
		(cudadata.trans_problemsperblock*cudadata.trans_rows*cudadata.trans_rows))*sizeof(double) + cudadata.n*sizeof(int);

	cudadata.rhs_threadsperstid = 32;
	cudadata.rhs_problemsperblock = max(min<int>((6*1024 - n*sizeof(int))/(2*n*sizeof(double)+sizeof(int)), 192/cudadata.rhs_threadsperstid),1);
	cudadata.rhs_sharedmemsize = n*(sizeof(int)+2*cudadata.rhs_problemsperblock*sizeof(double))+cudadata.rhs_problemsperblock*sizeof(int);

}

void Scheduler::destroyCUDAData(T_CUDADATA & cudadata)
{
	cudaFree(cudadata.sdiag);
	cudaFree(cudadata.mapped_sdiag);
	cudaFree(cudadata.n_rhselim);
	cudaFree(cudadata.n_matrixelim);
	cudaFree(cudadata.n_backsubst);
	cudaFree(cudadata.n_nonzeros);
	cudaFree(cudadata.nonzeros);
	cudaFree(cudadata.mapped_rhselim);
	cudaFree(cudadata.rhselim);
	cudaFree(cudadata.mapped_matrixelim);
	cudaFree(cudadata.backsubst);

}



}